/*
 * Copyright (C) 1997-2008 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file server.c Paraslash's main server. */


/**
 * \mainpage Paraslash API Reference
 *
 * Starting points for getting an overview are
 *
 * probably:
 *
 * 	- The main programs: \ref server.c, \ref audiod.c, \ref client.c,
 *	  \ref audioc.c, \ref fsck.c,
 *	- Server: \ref server_command, \ref sender,
 *	- Audio file selector: \ref audio_format_handler, \ref mood, \ref afs_table,
 *	- Client: \ref receiver, \ref receiver_node, \ref filter, \ref filter_node.
 *
 *
 * The gory details, listed by topic:
 *
 * 	- Audio format handlers: \ref mp3_afh.c, \ref ogg_afh.c, \ref aac_afh.c,
 * 	- Decoders: \ref mp3dec.c, \ref oggdec.c, \ref aacdec.c,
 *	- Volume normalizer: \ref compress.c,
 *	- Output: \ref alsa_write.c, \ref osx_write.c,
 * 	- http: \ref http_recv.c, \ref http_send.c,
 *	- ortp: \ref ortp_recv.c, \ref ortp_send.c,
 *	- dccp: \ref dccp_recv.c, \ref dccp_send.c,
 *	- Audio file selector: \ref afs.c, \ref aft.c, \ref mood.c,
 *	- Afs structures: \ref afs_table, \ref audio_file_data,
 *	  \ref afs_info \ref afh_info,
 *	- Afs tables: \ref aft.c, \ref mood.c, \ref playlist.c,
 *	  \ref attribute.c, \ref score.c,
 *	- The virtual streaming system: \ref vss.c, \ref chunk_queue.c.
 *
 * Lower levels:
 *
 *	- Scheduling: \ref sched.c, \ref sched.h,
 * 	- Networking: \ref net.c,
 *	- File descriptors: \ref fd.c,
 *	- Signals: \ref signal.c,
 *	- Daemons: \ref daemon.c,
 *	- Strings: \ref string.c, \ref string.h,
 *	- Time: \ref time.c,
 *	- Spawning processes: \ref exec.c,
 *	- Inter process communication: \ref ipc.c,
 *	- The object storage layer: \ref osl.c,
 *	- Blob tables: \ref blob.c,
 *	- The error subssystem: \ref error.h.
 *	- Access control for paraslash senders: \ref acl.c, \ref acl.h.
 *
 * Low-level data structures:
 *
 *	- Doubly linked lists: \ref list.h,
 *	- Red-black trees: \ref rbtree.h, \ref rbtree.c,
 *	- Ring buffer: \ref ringbuffer.c, \ref ringbuffer.h,
 *	- Hashing: \ref hash.h, \ref sha1.h, \ref sha1.c,
 *	- Crypto: \ref crypt.c.
 *
 */

#include <signal.h>
#include <sys/types.h>
#include <dirent.h>

#include "para.h"
#include "error.h"
#include "server.cmdline.h"
#include "afh.h"
#include "string.h"
#include "afs.h"
#include "server.h"
#include "vss.h"
#include "config.h"
#include "close_on_fork.h"
#include "list.h"
#include "send.h"
#include "net.h"
#include "daemon.h"
#include "ipc.h"
#include "fd.h"
#include "sched.h"
#include "signal.h"
#include "user_list.h"

/** define the array of error lists needed by para_server */
INIT_SERVER_ERRLISTS;

/** shut down non-authorized connections after that many seconds */
#define ALARM_TIMEOUT 10

/**
 * Pointer to shared memory area for communication between para_server
 * and its children. Exported to vss.c. command.c and to afs.
 */
struct misc_meta_data *mmd;

/**
 * the configuration of para_server
 *
 * It also contains the options for the audio file selector, audio format
 * handler and all supported senders.
 */
struct server_args_info conf;

/** the file containing user information (public key, permissions) */
char *user_list_file = NULL;

extern void dccp_send_init(struct sender *);
extern void http_send_init(struct sender *);
extern void ortp_send_init(struct sender *);

/** the list of supported senders */
struct sender senders[] = {
	{
		.name = "http",
		.init = http_send_init,
	},
	{
		.name = "dccp",
		.init = dccp_send_init,
	},
#ifdef HAVE_ORTP
	{
		.name = "ortp",
		.init = ortp_send_init,
	},
#endif
	{
		.name = NULL,
	}
};


/* global variables for server-internal use */
static FILE *logfile;
static int mmd_mutex, mmd_shm_id;
static int signal_pipe;

/**
 * para_server's log function
 *
 * \param ll the log level
 * \param fmt the format string describing the log message
 */
void para_log(int ll, const char* fmt,...)
{
	va_list argp;
	FILE *outfd;
	struct tm *tm;
	time_t t1;
	char str[MAXLINE] = "";
	pid_t mypid;

	if (ll < conf.loglevel_arg)
		return;
	outfd = logfile? logfile : stderr;
	time(&t1);
	tm = localtime(&t1);
	strftime(str, MAXLINE, "%b %d %H:%M:%S", tm);
	fprintf(outfd, "%s ", str);
	if (conf.loglevel_arg <= INFO)
		fprintf(outfd, "%i: ", ll);
	mypid = getpid();
	if (conf.loglevel_arg <= INFO)
		fprintf(outfd, "(%d) ", (int)mypid);
	va_start(argp, fmt);
	vfprintf(outfd, fmt, argp);
	va_end(argp);
}

/*
 * setup shared memory area and get mutex for locking
 */
static void shm_init(void)
{
	void *shm;
	int ret = shm_new(sizeof(struct misc_meta_data));

	if (ret < 0)
		goto err_out;

	ret = shm_attach(ret, ATTACH_RW, &shm);
	if (ret < 0)
		goto err_out;
	mmd = shm;
	mmd_shm_id = ret;

	ret = mutex_new();
	if (ret < 0)
		goto err_out;
	mmd_mutex = ret;

	mmd->num_played = 0;
	mmd->num_commands = 0;
	mmd->events = 0;
	mmd->num_connects = 0;
	mmd->active_connections = 0;
	mmd->vss_status_flags = VSS_NEXT;
	mmd->new_vss_status_flags = VSS_NEXT;
	mmd->sender_cmd_data.cmd_num = -1;
	return;
err_out:
	PARA_EMERG_LOG("%s\n", para_strerror(-ret));
	exit(EXIT_FAILURE);
}

/**
 * lock the shared memory area containing the mmd struct
 *
 * \sa semop(2), struct misc_meta_data.
 */
void mmd_lock(void)
{
	mutex_lock(mmd_mutex);
}

/**
 * unlock the shared memory area containing the mmd struct
 *
 * \sa semop(2), struct misc_meta_data
 */

void mmd_unlock(void)
{
	mutex_unlock(mmd_mutex);
}

static void parse_config(int override)
{
	char *home = para_homedir();
	struct stat statbuf;
	int ret;
	char *cf;

	if (conf.config_file_given)
		cf = para_strdup(conf.config_file_arg);
	else
		cf = make_message("%s/.paraslash/server.conf", home);
	free(user_list_file);
	if (!conf.user_list_given)
		user_list_file = make_message("%s/.paraslash/server.users", home);
	else
		user_list_file = para_strdup(conf.user_list_arg);
	ret = stat(cf, &statbuf);
	if (ret && conf.config_file_given) {
		ret = -1;
		PARA_EMERG_LOG("can not stat config file %s\n", cf);
		goto out;
	}
	if (!ret) {
		int tmp = conf.daemon_given;
		struct server_cmdline_parser_params params = {
			.override = override,
			.initialize = 0,
			.check_required = 0,
			.check_ambiguity = 0
		};
		server_cmdline_parser_config_file(cf, &conf, &params);
		conf.daemon_given = tmp;
	}
	/* logfile */
	if (!conf.logfile_given && conf.daemon_given) {
		ret = -1;
		PARA_EMERG_LOG("fatal: daemon option, but no log file given\n");
		goto out;
	}
	if (conf.logfile_given)
		logfile = open_log(conf.logfile_arg);
	ret = 1;
out:
	free(cf);
	free(home);
	if (ret > 0)
		return;
	free(user_list_file);
	user_list_file = NULL;
	exit(EXIT_FAILURE);
}

static void setup_signal_handling(void)
{
	signal_pipe = para_signal_init(); /* always successful */

	PARA_NOTICE_LOG("setting up signal handlers\n");
	if (para_install_sighandler(SIGINT) < 0)
		goto err;
	if (para_install_sighandler(SIGTERM) < 0)
		goto err;
	if (para_install_sighandler(SIGHUP) < 0)
		goto err;
	if (para_install_sighandler(SIGCHLD) < 0)
		goto err;
	if (signal(SIGPIPE, SIG_IGN) == SIG_ERR)
		goto err;
	if (signal(SIGUSR1, SIG_IGN) == SIG_ERR)
		goto err;
	add_close_on_fork_list(signal_pipe);
	return;
err:
	PARA_EMERG_LOG("could not install signal handlers\n");
	exit(EXIT_FAILURE);
}

static unsigned init_network(void)
{
	int fd, ret = para_listen(AF_UNSPEC, IPPROTO_TCP, conf.port_arg);

	if (ret < 0)
		goto err;
	fd = ret;
	ret = mark_fd_nonblocking(fd);
	if (ret < 0)
		goto err;
	add_close_on_fork_list(fd); /* child doesn't need the listener */
	return fd;
err:
	PARA_EMERG_LOG("%s\n", para_strerror(-ret));
	exit(EXIT_FAILURE);
}

static void init_random_seed(void)
{
	unsigned int seed;
	int fd, ret = para_open("/dev/urandom", O_RDONLY, 0);

	if (ret < 0)
		goto err;
	fd = ret;
	ret = read(fd, &seed, sizeof(seed));
	if (ret < 0) {
		ret = -ERRNO_TO_PARA_ERROR(errno);
		goto out;
	}
	if (ret != sizeof(seed)) {
		ret = -ERRNO_TO_PARA_ERROR(EIO);
		goto out;
	}
	srandom(seed);
	ret = 1;
out:
	close(fd);
	if (ret >= 0)
		return;
err:
	PARA_EMERG_LOG("can not seed pseudo random number generator: %s\n",
		para_strerror(-ret));
	exit(EXIT_FAILURE);
}

uint32_t afs_socket_cookie;
int afs_socket;
static pid_t afs_pid;

static void init_afs(void)
{
	int ret, afs_server_socket[2];

	ret = socketpair(PF_UNIX, SOCK_DGRAM, 0, afs_server_socket);
	if (ret < 0)
		exit(EXIT_FAILURE);
	afs_socket_cookie = para_random((uint32_t)-1);
	afs_pid = fork();
	if (afs_pid < 0)
		exit(EXIT_FAILURE);
	if (!afs_pid) { /* child (afs) */
		close(afs_server_socket[0]);
		afs_init(afs_socket_cookie, afs_server_socket[1]);
	}
	close(afs_server_socket[1]);
	afs_socket = afs_server_socket[0];
	ret = mark_fd_nonblocking(afs_socket);
	if (ret < 0)
		exit(EXIT_FAILURE);
	add_close_on_fork_list(afs_socket);
	PARA_INFO_LOG("afs_socket: %d, afs_socket_cookie: %u\n", afs_socket,
		(unsigned) afs_socket_cookie);
}

static unsigned server_init(int argc, char **argv)
{
	/* connector's address information */
	int sockfd;

	init_random_seed();
	/* parse command line options */
	server_cmdline_parser(argc, argv, &conf);
	HANDLE_VERSION_FLAG("server", conf);
	para_drop_privileges(conf.user_arg, conf.group_arg);
	/* parse config file, open log and set defaults */
	parse_config(0);
	log_welcome("para_server", conf.loglevel_arg);
	shm_init(); /* init mmd struct */
	server_uptime(UPTIME_SET); /* reset server uptime */
	init_user_list(user_list_file);
	/* become daemon */
	if (conf.daemon_given)
		daemon_init();
	PARA_NOTICE_LOG("initializing audio format handlers\n");
	afh_init();
	PARA_NOTICE_LOG("initializing virtual streaming system\n");
	mmd->server_pid = getpid();
	setup_signal_handling();
	PARA_NOTICE_LOG("initializing the audio file selector\n");
	init_afs();
	vss_init();
	mmd_lock();
	/* init network socket */
	PARA_NOTICE_LOG("initializing tcp command socket\n");
	sockfd = init_network();
	PARA_NOTICE_LOG("server init complete\n");
	return sockfd;
}

/*
 * called when server gets SIGHUP or when client invokes hup command.
 */
static void handle_sighup(void)
{
	PARA_NOTICE_LOG("SIGHUP\n");
	close_log(logfile); /* gets reopened if necessary by parse_config */
	logfile = NULL;
	parse_config(1); /* reopens log */
	init_user_list(user_list_file); /* reload user list */
	if (afs_pid)
		kill(afs_pid, SIGHUP);
}

static void status_refresh(void)
{
	static int prev_uptime = -1, prev_events = -1;
	int uptime = server_uptime(UPTIME_GET), ret = 1;

	if (prev_events != mmd->events)
		goto out;
	if (mmd->new_vss_status_flags != mmd->vss_status_flags)
		goto out;
	if (uptime / 60 != prev_uptime / 60)
		goto out;
	ret = 0;
out:
	prev_uptime = uptime;
	prev_events = mmd->events;
	mmd->vss_status_flags = mmd->new_vss_status_flags;
	if (ret) {
		PARA_DEBUG_LOG("%d events, forcing status update\n",
			mmd->events);
		killpg(0, SIGUSR1);
	}
}

/**
 * the main function of para_server
 *
 * \param argc usual argument count
 * \param argv usual argument vector
 *
 * \return EXIT_SUCCESS or EXIT_FAILURE
 *
 */
int main(int argc, char *argv[])
{
	/* listen on sock_fd, new connection on new_fd */
	int sockfd, new_fd;
	char *peer_name;
	int i, max_fileno, ret;
	pid_t chld_pid;
	fd_set rfds, wfds;
	struct timeval *timeout;

	valid_fd_012();
	sockfd = server_init(argc, argv);
repeat:
	FD_ZERO(&rfds);
	FD_ZERO(&wfds);
	max_fileno = -1;
	/* check socket and signal pipe in any case */
	para_fd_set(sockfd, &rfds, &max_fileno);
	para_fd_set(signal_pipe, &rfds, &max_fileno);
	timeout = vss_preselect(&rfds, &wfds, &max_fileno);
	status_refresh();
	for (i = 0; senders[i].name; i++) {
		if (!senders[i].pre_select)
			continue;
		senders[i].pre_select(&max_fileno, &rfds, &wfds);
	}
	mmd_unlock();
	ret = para_select(max_fileno + 1, &rfds, &wfds, timeout);
	mmd_lock();
	vss_post_select(&rfds, &wfds);
	if (ret < 0)
		goto repeat;
	for (i = 0; senders[i].name; i++) {
		if (!senders[i].post_select)
			continue;
		senders[i].post_select(&rfds, &wfds);
	}
	vss_send_chunk();
	status_refresh();
	if (FD_ISSET(signal_pipe, &rfds)) {
		int sig;
		pid_t pid;
		sig = para_next_signal();
		switch (sig) {
		case SIGHUP:
			handle_sighup();
			break;
		case SIGCHLD:
			for (;;) {
				ret = para_reap_child(&pid);
				if (ret <= 0)
					break;
				if (pid != afs_pid)
					continue;
				PARA_EMERG_LOG("fatal: afs died\n");
				goto genocide;
			}
			break;
		/* die on sigint/sigterm. Kill all children too. */
		case SIGINT:
		case SIGTERM:
			PARA_EMERG_LOG("terminating on signal %d\n", sig);
genocide:
			kill(0, SIGTERM);
			mutex_destroy(mmd_mutex);
			shm_detach(mmd);
			shm_destroy(mmd_shm_id);

			exit(EXIT_FAILURE);
		}
	}
	if (mmd->sender_cmd_data.cmd_num >= 0) {
		int num = mmd->sender_cmd_data.cmd_num,
			s = mmd->sender_cmd_data.sender_num;

		if (senders[s].client_cmds[num])
			senders[s].client_cmds[num](&mmd->sender_cmd_data);
		mmd->sender_cmd_data.cmd_num = -1;
	}
	if (!FD_ISSET(sockfd, &rfds))
		goto repeat;

	new_fd = para_accept(sockfd, NULL, 0);
	if (new_fd < 0)
		goto repeat;
	peer_name = remote_name(new_fd);
	PARA_INFO_LOG("got connection from %s, forking\n", peer_name);
	mmd->num_connects++;
	mmd->active_connections++;
	random();
	chld_pid = fork();
	if (chld_pid < 0) {
		PARA_CRIT_LOG("fork failed\n");
		goto repeat;
	}
	if (chld_pid) {
		close(new_fd);
		/* parent keeps accepting connections */
		goto repeat;
	}
	alarm(ALARM_TIMEOUT);
	close_listed_fds();
	para_signal_shutdown();
	/*
	 * put info on who we are serving into argv[0] to make
	 * client ip visible in top/ps
	 */
	for (i = argc - 1; i >= 0; i--)
		memset(argv[i], 0, strlen(argv[i]));
	sprintf(argv[0], "para_server (serving %s)", peer_name);
	return handle_connect(new_fd, peer_name);
}
