/*
 * Copyright (C) 2006-2008 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file stdout.c Functions that deal with writing to stdout. */

#include <dirent.h> /* readdir() */
#include <assert.h>

#include "para.h"
#include "string.h"
#include "list.h"
#include "sched.h"
#include "fd.h"
#include "error.h"
#include "stdout.h"

/**
 * The pre_select function of the stdout task.
 *
 * \param s The scheduler this task was registered to.
 * \param t The task structure of the stdout task.
 *
 * This function is always successful. If there is data available in the input
 * buffer, it adds \p STDOUT_FILENO to the write fd set of \a s.
 */
static void stdout_pre_select(struct sched *s, struct task *t)
{
	struct stdout_task *sot = t->private_data;

	t->ret = 1;
	sot->check_fd = 0;
	if (!*sot->loaded) {
		if (*sot->input_error) {
			t->ret = *sot->input_error;
			s->timeout.tv_sec = 0;
			s->timeout.tv_usec = 1;
		}
		return;
	}
	sot->check_fd = 1;
	para_fd_set(STDOUT_FILENO, &s->wfds, &s->max_fileno);
}

/**
 * The post select function of the stdout task.
 *
 * \param s The scheduler this task was registered to.
 * \param t The task structure of the stdout task.
 *
 * This function checks if \p STDOUT_FILENO was included by in the write fd set
 * of \a s during the previous pre_select call.  If yes, and \p STDOUT_FILENO
 * appeears to be writable, the data loaded in the input buffer is written to
 * stdout.
 */
static void stdout_post_select(struct sched *s, struct task *t)
{
	struct stdout_task *sot = t->private_data;
	ssize_t ret;

	t->ret = 1;
	if (!sot->check_fd) {
		if (*sot->input_error)
			t->ret = *sot->input_error;
		return;
	}
	if (!FD_ISSET(STDOUT_FILENO, &s->wfds))
		return;
	t->ret = -E_STDOUT_WRITE;
	ret = write(STDOUT_FILENO, sot->buf, *sot->loaded);
	if (ret <= 0)
		return;
	*sot->loaded -= ret;
	if (*sot->loaded)
		memmove(sot->buf, sot->buf + ret, *sot->loaded);
	t->ret = 1;
}

static void stdout_default_event_handler(struct task *t)
{
	PARA_NOTICE_LOG("%p: %s\n", t, para_strerror(-t->ret));
	unregister_task(t);
}

/**
 * Initialize a stdout task structure with default values.
 *
 * \param sot The stdout task structure.
 *
 * This fills in the pre/post select function poinzters of the task structure
 * given by \a sot. It also sets up a default error handler which unregisters
 * the task on errors and clears the eof flag of \a sot.
 */
void stdout_set_defaults(struct stdout_task *sot)
{
	int ret;

	sot->task.private_data = sot;
	sot->task.pre_select = stdout_pre_select;
	sot->task.post_select = stdout_post_select;
	sot->task.event_handler = stdout_default_event_handler;
	sot->error = 0;
	sprintf(sot->task.status, "stdout writer");
	ret = mark_fd_nonblocking(STDOUT_FILENO);
	if (ret >= 0)
		return;
	PARA_EMERG_LOG("%s\n", para_strerror(-ret));
	exit(EXIT_FAILURE);
}
